#include <stdio.h>
#include <io.h>
#include <fcntl.h>

struct WAB_CNK {
	unsigned char name[4];
	unsigned long size;
} WCNK;

struct WAB_FMT {
	unsigned short id;      // tH[}bgID
	unsigned short ch_cnt;  // `l
	unsigned long  hz;      // TvOg
	unsigned long  spd;     // σf[^x
	unsigned short blk_siz; // ubNTCY
	unsigned short bit_siz; // PTṽrbg
} WFMT;

unsigned char TMP0[0x4000];
unsigned char TMP1[0x4000];

unsigned short DES_OFF;
unsigned char  DES_BIT;
unsigned short DES_LIM;

void sub_init(unsigned short lim)
{
	DES_OFF=0;
	DES_BIT=0x80;
	DES_LIM=lim;
	memset(TMP1,0,0x4000);
}

void bit_disp(unsigned short cnt)
{
	unsigned short off;
	unsigned char  bit;

	off=0; bit=0x80;
	for(;;){
		if(TMP1[off]&bit) printf("|");
		else              printf("_");
		bit>>=1;
		if(!bit){
			bit=0x80; off++;
			if(!(off%8)) printf("\n");
			if(off>=cnt) break;
		}
	}
}

short sub_conv(FILE *des,short cnt)
{
	short i,n;

	for(i=0; i<cnt; i++){
		if(!DES_BIT){
			DES_BIT=0x80; DES_OFF++;
			if(DES_OFF>=0x4000){
				if(des!=NULL){
					fwrite(TMP1,1,DES_OFF,des);
					if(ferror(des)) return 1;
				} else bit_disp(DES_OFF);
				DES_OFF=0; memset(TMP1,0,0x4000);
			}
		}

		n=(short)TMP0[i];
		n-=127; if(n>=0) n++;
		n/=4; n*=4;
		if(n>DES_LIM) TMP1[DES_OFF]|=DES_BIT;
		DES_BIT>>=1;
	}

	return 0;
}

short sub_term(FILE *des)
{
	if(des==NULL){
		bit_disp(DES_OFF);
		printf("\n");
		return 0;
	}
	fwrite(TMP1,1,DES_OFF,des);
	return (ferror(des)?1:0);
}

short chk(unsigned char *str,unsigned char *buf)
{
	unsigned char i;
	for(i=0; str[i]; i++){
		if(str[i]!=buf[i]) return 1;
	}
	return 0;
}

short wav2bit(short max,unsigned char *wav_name,unsigned char *tap_name)
{
	long next,flng,lng;
	short cnt,no,err;
	FILE *src;
	FILE *des;

	src=des=NULL;
	sub_init(max);

	no=1; err=0;
	while(!err){
		switch(no){
			case 1:
				src=fopen(wav_name,"rb");
				if(src==NULL) err=1;
			break;
			case 2:
				if(tap_name!=NULL){
					des=fopen(tap_name,"wb");
					if(des==NULL){ err=2; break; }
					fwrite(&WFMT.hz,1,4,des); // <- а
				}
			break;
			case 3:
				fread(&WCNK,1,8,src);
				if(ferror(src)){ err=1; break; }
				if(chk("RIFF",WCNK.name)){ err=3; break; }
				flng=ftell(src)+WCNK.size;
			break;
			case 4:
				fread(TMP0,1,4,src);
				if(ferror(src)){ err=1; break; }
				if(chk("WAVE",TMP0)){ err=3; break; }
				next=ftell(src);
			break;
			case 5:
				while(next<flng && !err){
					fseek(src,next,SEEK_SET);
					if(fread(&WCNK,1,8,src)<8) break;
					if(feof(src)) break;
					if(ferror(src)){ err=1; break; }
					next=ftell(src)+WCNK.size;

					if(       !chk("fmt ",WCNK.name)){
						fread(&WFMT,1,16,src);
					} else if(!chk("data",WCNK.name)){
						lng=WCNK.size;
						while(lng>0l){
							cnt=(lng>0x4000l?0x4000:(short)lng);
							lng-=(long)fread(TMP0,1,cnt,src);
							if(ferror(src)){ err=1; break; }
							if(sub_conv(des,cnt)){ err=2; break; }
						}
					}
				}
			break;
		}
		no++; if(no>5) break;
	}
	sub_term(des);

	if(src!=NULL) fclose(src);
	if(des!=NULL){
		fclose(des);
		des=fopen(tap_name,"r+b");
			fseek(des,0l,SEEK_SET);
			fwrite(&WFMT.hz,1,4,des);
		fclose(des);
	}

	return err;
}

void main(int argc,unsigned char *argv[])
{
	unsigned char *src;
	unsigned char *des;
	unsigned short lim;
	short i,n,err;

	err=0; src=des=NULL; lim=128;
	for(i=1; i<argc && !err; i++){
		if(argv[i][0]=='/' || argv[i][0]=='-'){
			lim=atoi(&argv[i][1]);
		} else {
			if(     src==NULL) src=argv[i];
			else if(des==NULL) des=argv[i];
			else err=1;
		}
	}
	if(src==NULL || err){
		printf("v`us`oփRo[g\n");
		printf("  WAV2TAP [t@C.WAV] [t@C.TAP]\n");
		printf("v`ũeXg\n");
		printf("  WAV2TAP [t@C.WAV]\n");
		printf("̎w(ftHgPQW)\n");
		printf("  WAV2TAP /128 [t@C.WAV]\n");
		printf("/*************************************************\n");
		printf("Wrbgk̂v`ut@Cw肵ĉ    \n");
		printf("TvO[ǵAȂׂW̔{ɂĉ\n");
		printf("*************************************************/\n");
		exit(1);
	}

	if(wav2bit(lim,src,des)) printf("Error!\n");
	else                     printf("Ok.\n");
}

