#include <stdio.h>
#include <io.h>
#include <dos.h>
#include <fcntl.h>
#include <i86.h>
#include "x1.h"
#include "x1_crtc.h"
#include "x1_pcg.h"
#include "x1_vram.h"

/**********************************************************************
	AhX̌vZ (X1)
**********************************************************************/

word jis2adr_x1(word jis)
{
	word jh,jl,adr;

	jh=(jis>>  8);
	jl=(jis&0xff);
	if(jh>0x28) adr=0x4000+((jh-0x30)*0x600);
	else        adr=0x0100+((jh-0x21)*0x600);
	if(jl>=0x20) adr+=((jl-0x20)*0x10);

	return adr;
}

word adr2jis_x1(word adr)
{
	word jh,jl;

	if(adr<0x4000){ jh=adr-0x0100; jh=0x21+(jh/0x600); }
	else{           jh=adr-0x4000; jh=0x30+(jh/0x600); }
	if(jh>0x28) adr-=(0x4000+((jh-0x30)*0x600));
	else        adr-=(0x0100+((jh-0x21)*0x600));
	jl=0x20; if(adr) jl+=(adr/0x10);

	return ((jh<<8)|jl);
}

/**********************************************************************
	AhX̌vZ (X1Turbo)
**********************************************************************/

word jis2adr_x1t(word jis)
{
	word rw1,rw2,rw3;
	word rh,rl;

	if(!jis) return 0;

	rl =(jis&0x1f);
	rl|=(jis/8)&0xe0;
	rw1=(jis/0x1000)&7; // jh6,5,4
	rw2=(jis/0x0800)&1; // jh3
	rw3=(jis/0x0020)&3; // jl6,5
	switch(rw1){
		case 0: case 1: rh=rl=0; break;
		case 2:
			rh=rw2;
			if(     rw3==2) rh|=2;
			else if(rw3==3) rh^=1;
		break;
		case 3: case 4: case 5: case 6:
			rh =(rw1-3)*3;
			rh+=(rw3+1);
			rh=(rh*2)|rw2;
		break;
		case 7:
			rh=0x1c|rw2;
			if(rw3==2) rh|=2;
			if(rw3==3) rh^=1;
		break;
	}

	return (rh<<8)|rl;
}

word adr2jis_x1t(word adr)
{
	word j1,j2;
	word rl,rh;

	rh=(adr>> 8);
	rl=(adr&255);

	rh&=0x1f;
	if(!rl && !rh) return 0;

	j2=(rl&0x1f); // rl4,3,2,1,0
	j1=(rl/0x20)&7; // rl7,6,5

	if(     rh<0x04){ // 2121-277e
		j1|=0x20;
		switch(rh&3){
			case 0: j2|=0x20; break;
			case 1: j2|=0x60; break;
			case 2: j2|=0x40; break;
			default: j1=j2=0; break;
		}
	} else if(rh>0x1c){ // 7021-777e
		j1|=0x70;
		switch(rh&3){
			case 0: j2|=0x20; break;
			case 1: j2|=0x60; break;
			case 2: j2|=0x40; break;
			default: j1=j2=0; break;
		}
	} else {
		j1|=((((rh>>1)+7)/3)  )*0x10;
		j1|=(rh&1)*8;
		j2|=((((rh>>1)+1)%3)+1)*0x20;
	}

	return (j1<<8)|j2;
}

/**********************************************************************
	JIS/S-JIS̕ϊ
**********************************************************************/

word sjis2jis(word sjis)
{
	word c1,c2;

	if(!sjis) return 0;

	c1=(sjis>> 8);
	c2=(sjis&255);

	if(     c1<0xa0) c1-=0x70;
	else if(c1<0xf0) c1-=0xb0;
	else return sjis;
	if(c2>=0x80) c2--;
	c1*=2;
	if(c2>=0x9e) c2-=0x5e;
	else c1--;
	c2-=0x1f;

	return ((c1<<8)|c2);
}

word jis2sjis(word jis)
{
	word c1,c2;

	if(!jis) return 0;

	c1=(jis>> 8);
	c2=(jis&255);

	if(c1&1){
		c2+=0x1f;
		if(c2>=0x7f) c2++;
	} else {
		c2+=0x7e;
	}
	c1=(c1-0x20-1)/2+0x81;
	if(c1>=0xa0) c1+=0x40;

	return ((c1<<8)|c2);
}

/**********************************************************************
	ihrR[h犿tHgp^[Ԃ
**********************************************************************/

void knj_ptn(word jis,byte *kl,byte *kr)
{
	byte *fnt;
	word sjis,i;

	sjis=jis2sjis(jis);
	if(sjis<0x100){
		fnt=KNJ_FNT+((long)sjis*16l);
	} else if(sjis>=0x8140 && sjis<0x84c0){
		fnt=KNJ_FNT+(0x01000l+(long)(sjis-0x8140)*32l);
	} else if(sjis>=0x8890 && sjis<0xa000){
		fnt=KNJ_FNT+(0x08000l+(long)(sjis-0x8890)*32l);
	} else if(sjis>=0xe040 && sjis<0xeab0){
		fnt=KNJ_FNT+(0x36e00l+(long)(sjis-0xe040)*32l);
	} else {
		memset(kl,0,16);
		memset(kr,0,16);
		return;
	}

	for(i=0; i<16; i++){
		(*kl++)=(*fnt++);
		(*kr++)=(*fnt++);
	}
}

/***********************************************************************
	obf
***********************************************************************/

// bqsbQƒ̂u|q`lAhXԂ(PP)
// (ՂɂT񂫂)
word nowsyncoffset(void)
{
	word ret;

	// ret = ( ((V_CNT-CRT_YL)/FNT_YL) + TXT_YL) * TXT_XL;

	ret =(V_CNT/FNT_YL)*TXT_XL;
	ret+=(H_CNT*TXT_XL)/HDISP_CNT;
	ret+=VRAM_TOP;

	return (ret<=0x07ff?ret:0x07ff);
}

word pcg_adr2(word off)
{
	word i,adr;

	if(TXT_ATR[off]&0x20) return off;

	adr=(off<TXT_XL?0:(off-TXT_XL));
	for(i=0; i<TXT_XL; i++){
		if(TXT_ATR[adr]&0x20) return adr;
		adr++;
	}

	return off;
}

word chr_adr2(word off)
{
	word i,adr;

	if(!(TXT_ATR[off]&0x20)) return off;

	adr=(off<TXT_XL?0:(off-TXT_XL));
	for(i=0; i<TXT_XL; i++){
		if(!(TXT_ATR[adr]&0x20)) return adr;
		adr++;
	}

	return off;
}

word pcg_adr(void)
{
	if(TXT_ATR[0x7ff] & 0x20) return 0x7ff;
	if(TXT_ATR[0x3ff] & 0x20) return 0x3ff;
	if(TXT_ATR[0x5ff] & 0x20) return 0x5ff;
	if(TXT_ATR[0x1ff] & 0x20) return 0x1ff;
	return 0x3ff;
}

word chr_adr(void)
{
	if(!(TXT_ATR[0x7ff] & 0x20)) return 0x7ff;
	if(!(TXT_ATR[0x3ff] & 0x20)) return 0x3ff;
	if(!(TXT_ATR[0x5ff] & 0x20)) return 0x5ff;
	if(!(TXT_ATR[0x1ff] & 0x20)) return 0x1ff;
	return 0x3ff;
}

//----------------------------------------------------------------------

void x1_pcg_w(word Port,byte Value)
{
	static byte off1=0;
	static byte off2=0;
	static byte off4=0;
	word off;
	byte ank,knj;

	if(!VSYNC_FLG) off1=off2=off4=0;
	VSYNC_FLG=1;

	if(!PCG_MODE){
		off=nowsyncoffset();
		off=pcg_adr2(off);
		ank=(word)TXT_ANK[off];
		knj=(word)TXT_KAN[off];
		switch(Port>>8){
			case 0x15: PCG_B[ank][off1++]=Value; off1%=8; break;
			case 0x16: PCG_R[ank][off2++]=Value; off2%=8; break;
			case 0x17: PCG_G[ank][off4++]=Value; off4%=8; break;
		}
	} else {
		off=pcg_adr();
		ank=(word)TXT_ANK[off];
		knj=(word)TXT_KAN[off];
		off=(Port&15);
		if(knj&0x90){ ank&=0xfe; if(off&1) ank++; }
		off>>=1;
		switch(Port>>8){
			case 0x15: PCG_B[ank][off]=Value; break;
			case 0x16: PCG_R[ank][off]=Value; break;
			case 0x17: PCG_G[ank][off]=Value; break;
		}
	}
}

byte x1_pcg_r(word Port)
{
	static byte off0=0;
	static byte off1=0;
	static byte off2=0;
	static byte off4=0;
	static word vjis=0;
	static byte kl[16];
	static byte kr[16];
	word ank,knj;
	word h,l,n,off;
	byte val,dmy;

	val=0xff;
	h=(Port>>8);
	l=(Port&15);

	if(!VSYNC_FLG) off1=off2=off4=0;
	VSYNC_FLG=1;

	if(!PCG_MODE){
		off=nowsyncoffset();
		if(h==0x14) off=chr_adr2(off);
		else        off=pcg_adr2(off);
		ank=(word)TXT_ANK[off];
		knj=(word)TXT_KAN[off];
		switch(h){
			case 0x14: val=ANK_FNT[ank][off0++]; off0%=8; break;
			case 0x15: val=PCG_B[ank][off1++]; off1%=8; break;
			case 0x16: val=PCG_R[ank][off2++]; off2%=8; break;
			case 0x17: val=PCG_G[ank][off4++]; off4%=8; break;
		}
	} else if(h==0x14){
		off=chr_adr();
		ank=(word)TXT_ANK[off];
		knj=(word)TXT_KAN[off];
		if(knj & 0x80){
			n=(knj<<8)|ank;
			if(vjis!=n){ vjis=n; knj_ptn(adr2jis_x1t(vjis),kl,kr); }
			if(knj&0x40) val=kr[l];
			else         val=kl[l];
		} else if(CRTC_REG.dsp_flg & 0x40){ // 16׽
			val=KNJ_FNT[(ank*16)+l];
		} else { // 8׽
			val=ANK_FNT[ank][l>>1];
		}
	} else {
		off=pcg_adr();
		ank=(word)TXT_ANK[off];
		knj=(word)TXT_KAN[off];
		if(knj&0x90){ ank&=0xfe; if(l&1) ank++; }
		l>>=1;
		switch(h){
			case 0x15: val=PCG_B[ank][l]; break;
			case 0x16: val=PCG_R[ank][l]; break;
			case 0x17: val=PCG_G[ank][l]; break;
		}
	}

	return val;
}

/***********************************************************************
	ǂݍ݃|[g (X1)
***********************************************************************/

static word KC_H=0;
static word KC_L=0;
static word K_CNT=0;
static word K_FLG=0;
static byte KR_L[16]={0};
static byte KR_R[16]={0};

void x1_knj_w(word Port,byte Value)
{
	word adr,jis;

	Port&=15;
	if(     Port==0) KC_L=(word)Value;
	else if(Port==1) KC_H=(word)Value;
	else if(Port==2){
		// K_CNT=K_FLG=0;
		adr =KC_H; adr<<=8;
		adr|=KC_L&0xf0; jis=adr2jis_x1(adr);
		knj_ptn(jis,KR_L,KR_R);
	}
}

byte x1_knj_r(word Port)
{
	word val;

	Port&=15;
	val=0xff;

	if(KC_H){
		if(     Port==0){ val=KR_L[K_CNT]; K_FLG|=1; }
		else if(Port==1){ val=KR_R[K_CNT]; K_FLG|=2; }
		if(K_FLG==3){ K_FLG=0; K_CNT++; K_CNT%=16; }
	} else {
		if(     Port==0){
			val=(jis2adr_x1(KC_L<<8)>>8);
		} else if(Port==1){
			val=0;
		}
	}

	return (byte)val;
}

