/*** Z80Em: Portable Z80 emulator *******************************************/
/***                                                                      ***/
/***                                 Z80.h                                ***/
/***                                                                      ***/
/*** This file contains the function prototypes and variable declarations ***/
/***                                                                      ***/
/*** Copyright (C) Marcel de Kogel 1996,1997                              ***/
/***     You are not allowed to distribute this software commercially     ***/
/***     Please, notify me, if you make any changes to this file          ***/
/****************************************************************************/

#ifndef NULL
#define	NULL	((void *)0)
#endif

#define		FALSE		0
#define		TRUE		1

typedef unsigned char   byte;
typedef unsigned short  word;
typedef unsigned long  dword;
#define	Z80OPCODE static void

#define S_FLAG          0x80
#define Z_FLAG          0x40
#define H_FLAG          0x10
#define V_FLAG          0x04
#define N_FLAG          0x02
#define C_FLAG          0x01

typedef union {
	struct { byte l,h,h2,h3; } B;
	struct { word l,h; } W;
	word D;
} z80_pair;

typedef struct {
	z80_pair	AF,BC,DE,HL,IX,IY,PC,SP;	/* -NS- */
	z80_pair	AF2,BC2,DE2,HL2;			/* -NS- */
	int		IFF1,IFF2,HALT,IM,I,R,R2;
	dword	pending_irq;
	int		pending_nmi;
} Z80_Regs;

//**********************************************************************

#define Z80_IGNORE_INT  -1
#define Z80_NMI_INT     -2

extern byte Z80_Running; // When 0, emulation terminates

//**********************************************************************

byte Z80_In(word port);
void Z80_Out(word port, byte value);

#define DoIn(lo,hi)		Z80_In((word)(lo)+((word)(hi)<<8))
#define DoOut(lo,hi,v)	Z80_Out((word)(lo)+((word)(hi)<<8), v)
#define DoIn_N			Z80_In(R.BC.D)
#define DoOut_N(v)		Z80_Out(R.BC.D,v)


byte Z80_RDMEM(word adrs);
void Z80_WRMEM(word adrs, byte value);

#define	Z80_RDOP(A)		Z80_RDMEM(A)
#define	Z80_RDOP_ARG(A)	Z80_RDMEM(A)
#define	M_RDMEM(A)		Z80_RDMEM(A)
#define	M_RDOP(A)		Z80_RDMEM(A)
#define	M_RDOP_ARG(A)	Z80_RDMEM(A)
#define	M_WRMEM(A,V)	Z80_WRMEM(A,V)

//**********************************************************************

word Z80_GetPC(void);
void Z80_SetPC(word adrs); // KM 99'06/29
word Z80_IFF2(void); // KM 99'06/29
void Z80_GetRegs(Z80_Regs *Regs);
void Z80_SetRegs(Z80_Regs *Regs);

void Z80_Cause_Interrupt(dword type);
void Z80_Clear_Pending_Interrupts(void);

void Z80_Reset(void);
word Z80_Execute(void);

word cbw(byte code);

